{-# LANGUAGE FlexibleContexts #-}
{-# LANGUAGE ScopedTypeVariables #-}
-- | This module contains the parsers
-- needed for term rewriting the tool.
module AutoStrat.Parsing
    (
    -- * Functions
      parseTerm
    , parseProb
    -- ** For Annotations
    , Problem(..)
    -- * Reexport
    , ParseError
    , SourcePos
    , errorPos
    , messageString
    , errorMessages
    ) where

import Data.Char (isSpace)
import Data.Functor.Identity
import Data.List

import Data.Rewriting.Annotation (NamedRule(..), AnnotElem(..), Annotation(..))
import Data.Rewriting.Rule.Type hiding (map)
import Data.Rewriting.Term.Parse hiding (parse)
import Data.Rewriting.Term.Type hiding (map, fold)

import Text.Parsec
import Text.Parsec.Char
import Text.Parsec.Error
import Text.Parsec.Prim

data Problem = Problem { vars  :: [String]
                       , trs   :: [NamedRule String String]
                       , annot :: Annotation String String
                       }
                       deriving Show

data ParseState = ParseST { tVars    :: [String]
                          , tTrs     :: [NamedRule String String]
                          , tAnnot   :: [(String, [TempAnnotElem])]
                          , ruleNum  :: Int
                          , ruleNames :: [String]
                          }

initState :: ParseState
initState = ParseST { tVars = []
                    , tTrs = []
                    , tAnnot = []
                    , ruleNum = 0
                    , ruleNames = []
                    }

freshRuleName :: Monad m => ParsecT s ParseState m String
freshRuleName = do
    st <- getState
    let i = ruleNum st
        nms = ruleNames st
        name = "rule" ++ show i
    modifyState $ \st -> st { ruleNum = i + 1 }
    if name `elem` nms
       then freshRuleName
       else return name

-- | @parseTerm vars inp@ parses Terms from @inp@,
-- where @vars@ are the defined variables.
--
-- >>> parseTerm ["x","y"] "f(g(x,y))"
-- Right (Fun "f" [Fun "g" [Var "x",Var "y"]])
--
parseTerm :: Stream s Identity Char
          => [String] -> s -> Either ParseError (Term String String)
parseTerm vars = parse (term vars) ""

-- | Parse a system, containing rewrite rules and annotations
--
-- __Examples__:
--
-- >>> trs <$> parseProb ["x"] "alpha: f(a,x) -> a A(f) = [1, alpha, 2]"
-- Right [NamedRule "alpha" (Rule {lhs = Fun "f" [Fun "a" [],Var "x"], rhs = Fun "a" []})]
--
-- >>> annot <$> parseProb ["x"] "alpha: f(a,x) -> a A(f) = [1, alpha, 2]"
-- Right [("f",[0,alpha,1])]
--
parseProb :: Stream s Identity Char => s -> Either ParseError Problem
parseProb inp = runParser (spaces *> parseMain) initState "MainInput" inp

-- | Parses the TRS data into a set of rules and Annotations.
parseMain :: Stream s m Char => ParsecT s ParseState m Problem
parseMain = do
    prob <- trsP
    eof
    case convProb prob of
      Right p -> return p
      Left e -> unexpected e


trsP :: Stream s m Char => ParsecT s ParseState m ParseState
trsP = optional varsP >> rulesP >> optional annotP >> getState

varsP :: Stream s m Char => ParsecT s ParseState m ()
varsP = decl "VAR" varsP' (\r p -> p { tVars = r })
    where
        varsP' = many (lexme $ ident "(),:")

rulesP :: Stream s m Char => ParsecT s ParseState m ()
rulesP = decl "RULES" namedRules (\r p -> p { tTrs = r })

annotP :: Stream s m Char => ParsecT s ParseState m ()
annotP = decl "ANNOTATION" annotation (\r p -> p { tAnnot = r })
    where
        annotation = many $ lexme annot'
        annot' = par $ do
            name <- lexme $ ident "(),:"
            es <- many annotElem
            return (name, es)

decl :: Stream s m Char
     => String -> ParsecT s u m a -> (a -> u -> u)
     -> ParsecT s u m ()
decl name p f = par $ do
    try (lexme $ string name) <?> (name ++ " block")
    r <- p
    modifyState $ f r

-- | @ruleP vars@ is a parser for a unnamed rule,
-- where @vars@ is the list of variable names.
ruleP :: Stream s m Char => [String] -> ParsecT s u m (Rule String String)
ruleP vars = Rule <$> term vars <* lexme (string "->") <*> term vars

-- | @namedRules vars@ parses the rules of a TRS with/without names,
-- where @vars@ is the list of variable names.
--
-- The names of the rules are used when rewriting with annotations.
namedRules :: Stream s m Char
           => ParsecT s ParseState m [NamedRule String String]
namedRules = do
    vs <- tVars <$> getState
    many (namedRule vs)

-- | @namedRule vars@ is a parser for a named rule,
-- where @vars@ is the list of variable names.
--
-- The names are used when rewriting with annotations.
namedRule :: Stream s m Char
          => [String] -> ParsecT s ParseState m (NamedRule String String)
namedRule vars = do
    n <- optionMaybe (try $ lexme (ident "(),:" <* char ':'))
    name <- case n of
      Nothing -> freshRuleName
      Just n' -> return n'
    rl <- ruleP vars
    return $ NamedRule name rl


convProb :: ParseState -> Either String Problem
convProb st = case mapM name2Rule annots of
                Left e -> Left e
                Right x -> Right  Problem { vars = tVars st
                                          , trs = rls
                                          , annot = x }
    where
        rls = tTrs st
        annots = tAnnot st
        lkpRules = map (\(NamedRule nm rl) -> (nm,rl)) rls
        name2Rule :: (String, [TempAnnotElem]) -> Either String (String, [AnnotElem String String])
        name2Rule (func, elems) = case mapM conv elems of
                                    Left e    -> Left e
                                    Right els -> Right (func, els)
        conv :: TempAnnotElem -> Either String (AnnotElem String String)
        conv (TempPos i) = Right $ Pos (i-1)
        conv (TempRule name) = case lookup name lkpRules of
                                 Nothing -> Left ("No rule with name: " ++ name)
                                 Just rl -> Right $ AnnotRule (NamedRule name rl)


-- | Temporary data structure for annotations.
data TempAnnotElem = TempPos Int
                   | TempRule String
    deriving Show

annotElem :: Stream s m Char
          => ParsecT s u m TempAnnotElem
annotElem = TempPos <$> index
        <|> TempRule <$> name
    where
        index = lexme (read <$> many1 digit)
        name = lexme (ident "(),:")

-- | @term@ is a term parser.
-- At the moment it's just Data.Rewriting.Term.Parse.parseWST .
term :: Stream s m Char => [String] -> ParsecT s u m (Term String String)
term = parseWST

-- Utils
lexme :: Stream s m Char => ParsecT s u m a -> ParsecT s u m a
lexme p = p <* spaces

symbol :: Stream s m Char => Char -> ParsecT s u m Char
symbol = lexme . char

-- | @ident taboo@ parses a non-empty sequence of non-space characters not
-- containing elements of @taboo@.
ident :: Stream s m Char => String -> ParsecT s u m String
ident tabooChars = try $ many1 (satisfy (\c -> not (isSpace c) && c `notElem` tabooChars))

par :: Stream s m Char => ParsecT s u m a -> ParsecT s u m a
par = between (lexme $ char '(') (lexme $ char ')')
