   {-# LANGUAGE ScopedTypeVariables #-}
-- | This module implements rewrite annotations as described by Van do Pol.
module Data.Rewriting.Annotation
    ( Annotation
    -- * Types
    , NamedRule (..)
    , getRule
    , AnnotElem (..)
    -- * Functions
    , normalize
    , normalizeIter
    , Zipper (..)
    , leftInnermostAnnotation
    , justInTimeAnnotation
    -- * Properties and Utils
    -- , isFull
    , neededPositions
    ) where

import Data.Rewriting.Context
import Data.Rewriting.Rule (Rule(..))
import qualified Data.Rewriting.Rules.Rewrite as R
import Data.Rewriting.Term hiding (map)
import Data.Rewriting.Pos

import Data.Maybe (fromMaybe, mapMaybe)
import Data.List (unfoldr, foldl', nub)
import qualified Data.Map.Strict as M
import qualified Data.Set as S

-- | An annotation maps a function symbol to a list of AnnotElems.
-- The AnnotElems tell the normalization procedure what to do.
type Annotation f v = [(f, [AnnotElem f v])]

-- | An AnnotElem can either be an Integer,
-- giving the position of the argument to be normalized,
-- or a rule to try and apply.
data AnnotElem f v
  = Pos Int                   -- ^ Position of the subterm to be normalized
  | AnnotRule (NamedRule f v) -- ^ Rule to be matched with the term
  deriving (Eq)

instance Show (AnnotElem f v) where
    show (Pos i) = show i
    show (AnnotRule (NamedRule name _)) = name

-- | Wraps the term-rewriting rules to allow them to be named,
-- for use in annotations.
data NamedRule f v
  = NamedRule { name :: String
              , rule :: Rule f v
              }
  deriving (Show, Eq)

getRule :: NamedRule f v -> Rule f v
getRule = rule

getName :: NamedRule f v -> String
getName = name


-- | @normalize annot term@ normalizes @term@ using the @annot@ annotation.
normalize :: (Ord v, Eq f) => Annotation f v -> Term f v -> Term f v
normalize annot v@(Var _)       = v
normalize annot term@(Fun rt _) = norm' term (fromMaybe [] $ lookup rt annot)
    where
        norm' term [] = term
        norm' term (Pos i:xs) =
            let normalizedSubterm = normalize annot `fmap` subtermAt term [i]
                term'             = replaceAt term [i] =<< normalizedSubterm
             in maybe term (`norm'` xs) term'
        norm' term (AnnotRule rl:xs) =
            case map R.result $ R.rootRewrite [getRule rl] term of
              []    -> norm' term xs
              (x:_) -> normalize annot x

lookupAnnot :: Eq f => Annotation f v -> Term f v -> [AnnotElem f v]
lookupAnnot _ (Var _) = []
lookupAnnot annot (Fun f _) = fromMaybe [] (lookup f annot)

-- Zipper
data Zipper f v = Zipper { stack :: Stack f v
                         , term :: Term f v
                         , annotation :: [AnnotElem f v]
                         } deriving Show
type Stack f v = [(Ctxt f v, [AnnotElem f v])]

initZipper :: (Eq f) => Annotation f v -> Term f v -> Zipper f v
initZipper annot term = Zipper { stack = []
                               , term = term
                               , annotation = lookupAnnot annot term
                               }

-- | Moves up the zipper structure
zipUp :: Zipper f v -> Zipper f v
zipUp zipper =
    case stack zipper of
      [] -> zipper
      ((context, parentAnnot):stack') ->
          Zipper { stack      = stack'
                 , term       = apply context (term zipper)
                 , annotation = parentAnnot
                 }

-- | Moves down the zipper structure at a given position
zipDown :: Int -> (Term f v -> [AnnotElem f v]) -> Zipper f v -> Maybe (Zipper f v)
zipDown p annotLookup zipper =
    case parentContext of
      Nothing   -> Nothing
      Just ctxt ->
          case newTerm of
            Nothing  -> Nothing
            Just trm -> Just Zipper { stack = (ctxt, oldAnnotation):oldStack
                                    , term = trm
                                    , annotation = annotLookup trm
                                    }
    where
        oldStack      = stack zipper
        oldAnnotation = annotation zipper
        oldTerm       = term zipper
        parentContext = ofTerm oldTerm [p]
        newTerm       = subtermAt oldTerm [p]

-- | Reconstructs term from current zipper state
getTerm :: Zipper f v -> Term f v
getTerm zipper =
    case stack zipper of
      [] -> term zipper
      _  -> getTerm (zipUp zipper)

-- | Pops the next annotation element of the annotation list.
-- Returns the to element and the new zipper.
popAnnot :: Zipper f v -> Maybe (AnnotElem f v, Zipper f v)
popAnnot zipper = case annotation zipper of
                    []     -> Nothing
                    (l:ls) -> Just (l, zipper { annotation = ls })


zippNorm :: (Eq f, Ord v) => Annotation f v -> Zipper f v -> Maybe (Zipper f v)
zippNorm annot zipper =
    case popAnnot zipper of
      Nothing ->                        -- No more Annotation elements
          if null (stack zipper) then Nothing else Just (zipUp zipper)
      Just (Pos i,zipper') ->           -- Normalize subterm at position i
          zipDown i (lookupAnnot annot) zipper'
      Just (AnnotRule nrl,zipper') ->   -- Apply rule nrl
          case R.rootRewrite [getRule nrl] (term zipper') of
            [] -> Just zipper'
            (res:_) -> Just zipper' { term = R.result res
                                    , annotation = lookupAnnot annot (R.result res)
                                    }
-- | Iterativly normalizes a term returning all intermediat Zipper states
normalizeIter :: (Eq f, Ord v) => Annotation f v -> Term f v -> [Zipper f v]
normalizeIter annot term = zipper : unfoldr zippNorm' zipper
    where
        zippNorm' = fmap (\x -> (x,x)) . zippNorm annot
        zipper = initZipper annot term

-- | Determines if an Annotation is Full given a TRS
--
-- >>> let rule = NamedRule "alpha" (Rule (Fun "f" [Fun "a" [], Var "x"]) (Var "x"))
-- >>> let annot = [("f", [Pos 0, AnnotRule rule, Pos 1])]
-- >>> isFull annot [rule]
-- True
isFull :: forall f v. (Ord f, Ord v)
       => Annotation f v -> [NamedRule f v] -> Bool
isFull annotation trs = allPositions && allRules
    where
        allRules :: Bool
        allRules = namesMap == rootToRulesMap

        allPositions :: Bool
        allPositions = M.foldrWithKey go True ruleArities
        annotationMap = M.fromList annotation

        namesMap :: M.Map f [NamedRule f v]
        namesMap = M.map filterName annotationMap

        positionsMap :: M.Map f [Int]
        positionsMap = M.map filterPos annotationMap

        filterPos [] = []
        filterPos (Pos i:xs) = i:filterPos xs
        filterPos (AnnotRule {}:xs) = filterPos xs
        filterName [] = []
        filterName (AnnotRule rl:xs) = rl:filterName xs
        filterName (Pos _:xs) = filterName xs

        go _ _ False = False
        go k arity True = case M.lookup k positionsMap of
                            Nothing -> False
                            Just pos -> S.fromList pos == S.fromList [0..arity-1]

        ruleArities :: M.Map f Int
        ruleArities = foldl' insertArity M.empty trs
        insertArity m (NamedRule _ rl) = case lhs rl of
                            Var x -> m
                            Fun f params -> M.insert f (length params) m

        rootToRulesMap :: M.Map f [NamedRule f v]
        rootToRulesMap = foldl' insertRule M.empty trs
        insertRule m namedRule@(NamedRule _ rl) = case lhs rl of
                            Var x -> m
                            (Fun f _) -> M.insertWith (++) f [namedRule] m


-- | Returns the needed positions of a rule.
--
-- Examples:
--
-- >>> neededPositions $ Rule { lhs = Fun "f" [Var "x", Var "x"], rhs = Var "x" }
-- [0,1]
--
-- >>> neededPositions $ Rule { lhs = Fun "f" [Fun "a" [], Var "x"], rhs = Var "x"}
-- [0]
neededPositions :: Ord v => Rule f v -> [Integer]
neededPositions rl =
    case lhs rl of
      Var _ -> []
      Fun f ls -> let varCount = foldl' addVar M.empty ls
                   in go varCount (zip ls [0..])
    where
        addVar mapping (Fun _ _ ) = mapping
        addVar mapping (Var v) = M.insertWith (+) v 1 mapping

        go count [] = []
        go count ((Fun _ _,i):xs) = i: go count xs
        go count ((Var v,i):xs) = if (count M.! v) > 1
                                       then i : go count xs
                                       else go count xs


-- | Returns the strategy annotation equivalent to the leftmost innermost strategy.
--
-- >>> leftInnermostAnnotation [NamedRule "decr" (Rule (Fun "f" [Var "x"]) (Var "x"))]
-- [("f",[0,decr])]
leftInnermostAnnotation :: forall f v. (Ord f, Ord v)
                     => [NamedRule f v] -> Annotation f v
leftInnermostAnnotation trs = M.toList . M.map mkAnnotList $ rootRuleMap
    where
        mkAnnotList :: [NamedRule f v] -> [AnnotElem f v]
        mkAnnotList = nub . concatMap mkAnnot'

        mkAnnot' :: NamedRule f v -> [AnnotElem f v]
        mkAnnot' namedRl@(NamedRule name rl) =
            case lhs rl of
              Var _ -> []
              Fun f args -> map Pos [0..length args - 1] ++ [AnnotRule namedRl]

        rootRuleMap :: M.Map f [NamedRule f v]
        rootRuleMap = foldl insertRule M.empty trs

        insertRule m namedRl@(NamedRule _ rl) =
            case lhs rl of
              Var _ -> m
              Fun f _ -> M.insertWith (++) f [namedRl] m

-- | Returns a Just-in-time strategy annotation.
-- This is a full and intime annotation for a given trs.
--
-- >>> let rule1 = NamedRule "alpha" (Rule (Fun "f" [Fun "a" [], Var "x"])(Fun "a" []))
-- >>> let rule2 = NamedRule "beta" (Rule (Fun "f" [Var "x", Var "x"]) (Var "x"))
-- >>> justInTimeAnnotation [rule1,rule2]
-- [("f",[0,alpha,1,beta])]
justInTimeAnnotation :: forall f v. (Ord f, Ord v)
                     => [NamedRule f v] -> Annotation f v
justInTimeAnnotation trs = filter (not . null . snd) . M.toList $ foldl' insertRule  positionMap trs
    where
        insertRule :: M.Map f [AnnotElem f v] -> NamedRule f v -> M.Map f [AnnotElem f v]
        insertRule m rl@(NamedRule name unrl) =
            case lhs unrl of
              Var _ -> m
              Fun f _ -> M.adjust (`insertInTime` rl) f m

        arityMap :: M.Map f Int
        arityMap = M.fromList $ lhsFuns ++ rhsFuns
            where
                unnamedRules = map getRule trs
                lhsFuns = concatMap (funs . withArity . lhs) unnamedRules
                rhsFuns = concatMap (funs . withArity . rhs) unnamedRules

        positionMap :: M.Map f [AnnotElem f v]
        positionMap = (\i -> map Pos [0..(i-1)]) <$> arityMap


insertInTime :: Ord v => [AnnotElem f v] -> NamedRule f v -> [AnnotElem f v]
insertInTime annot namedRl@(NamedRule name rl) =
    go (fromIntegral $ maximum (-1:needed)) (AnnotRule namedRl) annot
    where
        needed = neededPositions rl
        go p rl [] = [rl]
        go p rl (anrl@(AnnotRule _):xs) = anrl: go p rl xs
        go p rl (Pos i:xs)
          | i > p     = rl : Pos i : xs
          | otherwise = Pos i : go p rl xs

