-- | All properties of TRSs, based on a set of rules belong here.
module Data.Rewriting.Properties (
    -- * Types
    TRS,

    -- * Properties
    leftLinear,
    orthogonal,
    weaklyOrthogonal,
    leftNormal
    ) where

import Data.Rewriting.CriticalPair (left, right, cps')
import Data.Rewriting.Rule         (Rule, lhs)
import Data.Rewriting.Rules        (isLeftLinear)
import Data.Rewriting.Term hiding  (map, fold)

-- | A term-rewrite-system given by a list of rules.
type TRS f v = [Rule f v]


-- | @leftLinear rs@ is true iff all rules @rs@ are left-linear.
--
-- renamed from 'Data.Rewriting.Rules'.
leftLinear :: Ord v => TRS f v -> Bool
leftLinear = isLeftLinear


-- | A TRS is left normal if all left hand sides are normal.
-- A Term is normal if there is no function symbol
-- to the right of a variable.
leftNormal :: TRS f v -> Bool
leftNormal rs = all normal $ map lhs rs

normal :: Term f v -> Bool
normal t = all isVar . dropWhile isFun $ flatten t
    where
        flatten :: Term f v -> [Term f v]
        flatten t@(Var _) = [t]
        flatten (Fun f xs) = Fun f [] : concatMap flatten xs


-- | @orthogonal rs@ is @true@ iff the TRS defined by @rs@ is orthogonal.
--
-- A TRS is orthogonal if it is left-linear and has no critical-pairs.
orthogonal :: (Eq f, Ord v) => TRS f v -> Bool
orthogonal rs = isLeftLinear rs && null (cps' rs)


-- | @weaklyOrthogonal rs@ is @true@ iff
-- the TRS defined by @rs@ is weakly-orthogonal.
--
-- A TRS is weakly-orthogonal iff
-- it is left-linear and all critical-pairs are trivial.
-- A CP is trivial if overlapping steps get the same result.
weaklyOrthogonal :: (Eq f, Ord v) => TRS f v -> Bool
weaklyOrthogonal rs = isLeftLinear rs && all trivial (cps' rs)
    where
        trivial cp = left cp == right cp

