-- | This module contains all the strategies 
-- used by the tool.
module Data.Rewriting.Strategies (
    -- * Types
    TRS,
    -- * Strategies
    innerRewrite,
    outerRewrite,

    -- * Rewriting
    leftmostOuter,
    leftmostInner,
    maximalOuter,
    maximalInner,
    maximalStrat,
    rewriteFirst,
    rewriteAll,

    -- * Ops on TRSs
    definedFuns,
    -- * Predicates on Terms and TRSs
    normalForm
    ) where

import Data.Either (rights)
import Data.Maybe (fromMaybe, fromJust)
import Data.List (unfoldr)
import qualified Data.Map as M (map)

import Data.Rewriting.Rule.Type hiding (map)
import Data.Rewriting.Rules.Rewrite
import Data.Rewriting.Term.Ops
import Data.Rewriting.Substitution.Type
import Data.Rewriting.Substitution.Ops

-- | A term-rewrite-system given by a list of rules.
type TRS f v = [Rule f v]

type Strategy' f v v' = Term f v -> [(Term f v, [Reduct f v v'])]


-- | Returns the defined function symbols of the given rules.
definedFuns :: TRS f v -> [f]
definedFuns = rights . fmap (root . lhs)

-- | @normalForm rs t@ checks if @t@ is in normal form,
-- with regards to the rules @rs@
normalForm :: (Eq f, Eq v, Ord v') => TRS f v' -> Term f v -> Bool
normalForm rules term = null $ fullRewrite rules term



-- | Makes a rewrite function from a Strategy,
-- by always choosing the first reduct.
rewriteFirst :: Strategy f v v' -> Strategy' f v v'
rewriteFirst strat term =
    case strat term of
      [] -> [(term, [])]
      (reduct:_) -> (term, [reduct]) : rewriteFirst strat (result reduct)

-- rewriteFirst strat term = term : case strat term of
--                                    []    -> []
--                                    (x:_) -> rewriteFirst strat (result x)

-- | Rewrites all redexes given by the Strategy.
--
-- The Strategy has to be non overlapping. This is not checked.
rewriteAll :: Strategy f v v' -> Strategy' f v v'
rewriteAll strat = rewriteAll'
    where
        rewriteAll' term =
            case strat term of
              []      -> [(term, [])]
              reducts -> let term' = foldl mergeReducts term reducts
                          in (term,reducts) : rewriteAll' term'

        mergeReducts :: Term f v -> Reduct f v v' -> Term f v
        mergeReducts term reduct = fromMaybe term $ do
            let position = pos reduct
            subterm' <- subtermAt (result reduct) position
            replaceAt term position subterm'


-- Deterministic rewrite strategies

-- | Leftmost-outermost rewrite strategy
--
-- Rewrites leftmost outermost redex.
leftmostOuter :: (Eq f, Eq v, Ord v') => TRS f v' -> Strategy' f v v'
leftmostOuter = rewriteFirst . outerRewrite

-- | Leftmost-innermost rewrite strategy
--
-- Rewrites leftmost innermost redex.
leftmostInner :: (Eq f, Eq v, Ord v') => TRS f v' -> Strategy' f v v'
leftmostInner = rewriteFirst . innerRewrite

-- | Maximal-innermost rewrite strategy
--
-- Rewrites all innermost redexes in parallel.
maximalInner :: (Eq f, Eq v, Ord v') => TRS f v' -> Strategy' f v v'
maximalInner = rewriteAll . innerRewrite

-- | Maximal-outermost rewrite strategy
--
-- Rewrites all outermost redexes in parallel.
maximalOuter :: (Eq f, Eq v, Ord v') => TRS f v' -> Strategy' f v v'
maximalOuter = rewriteAll . outerRewrite

-- | Maximal rewrite strategy
--
-- Rewrites by performing maximal multisteps.
maximalStrat :: (Eq f, Ord v) => TRS f v -> Strategy' f v v
maximalStrat trs = maximalStrat'
    where
        isNormal = normalForm trs
        maximalStrat' term
          | isNormal term = [(term, [])]
          | otherwise = let (res,reduct) = maximalMultiStep trs term
                         in (term,reduct) : maximalStrat' res

-- | Performs a single maximal multipstep
--
--  If the term is in normal form it will perform the identity step.
maximalMultiStep :: (Eq f, Ord v) => TRS f v -> Term f v -> (Term f v, [Reduct f v v])
maximalMultiStep trs term@(Var _) = (term, [])
maximalMultiStep trs term@(Fun f ls) =
    case rootRewrite trs term of
      [] -> let params = map (maximalMultiStep trs) ls
                terms = map fst params
                reducts = concat [ [ r { pos = i : pos r } | r <- rs ]
                                 | (i,(_,rs)) <- zip [0..] params
                                 ]
             in (Fun f terms, reducts)
      (reduct:_) -> let sub' = mapSubst (fst . maximalMultiStep trs) $ subst reduct
                        term' = apply sub' (rhs . rule $ reduct)
                     in (term', [reduct])

-- | @mapSubst f sub@ applies @f@ to all Terms that @sub@ maps to.
--
-- So if @sub@ maps @Var "x"@ to a term @t@, then @mapSubst f sub@ maps @Var "x"@ to @f t@.
mapSubst :: (Term f v -> Term f v) -> Subst f v -> Subst f v
mapSubst f = fromMap . M.map f . toMap

