   {-# LANGUAGE OverloadedStrings #-}
module Controller.Annotations
    ( annotationController
    , suggestAnnotationController
    ) where

import Web.Scotty (ActionM)
import qualified Web.Scotty as S
import Data.Aeson
import Data.Aeson.Types
import Data.List (intercalate)

import qualified AutoStrat.Parsing as P
import Data.Rewriting.Annotation (Annotation, AnnotElem (..), NamedRule (..))
import qualified Data.Rewriting.Annotation as N
import Data.Rewriting.Context.Type

import Controller.Orphans () -- Import instances

data Request = Request
    { problem :: String
    , term    :: String
    } deriving (Show)

instance FromJSON Request where
    parseJSON (Object o) =
        Request <$> o .: "problem" <*> o .: "term"
    parseJSON invalid = typeMismatch "Annotation Request" invalid

annotationController :: ActionM ()
annotationController = do
    req <- S.jsonData
    case handleReq req of
      Right res -> S.json res
      Left err  -> S.json err

suggestAnnotationController :: ActionM ()
suggestAnnotationController = do
    req <- S.jsonData
    case suggestAnnotHandler req of
      Right res -> S.json res
      Left err -> S.json err

data Suggestion = Suggest { annot :: Annotation String String
                          }

instance ToJSON Suggestion where
    toJSON sugg = object [ "annotation" .= prettyAnnot (annot sugg) ]

prettyAnnot :: Annotation String String -> [String]
prettyAnnot ann = "(ANNOTATION" : map go ann ++ [")"]
    where
        go (f,ann) = "  (" ++ f ++ " "
                  ++ intercalate " " (map prettyAnnotElem ann)
                  ++ ")"
        prettyAnnotElem (Pos i) = show (i+1)
        prettyAnnotElem (AnnotRule rl) = name rl
        name (NamedRule nm _) = nm

suggestAnnotHandler :: Request -> Either ReqError Suggestion
suggestAnnotHandler req = do
    prob <- liftErr "main" $ P.parseProb (problem req)
    let res = Suggest { annot = N.justInTimeAnnotation (P.trs prob) }
    return res

data ReqError = ReqError { field :: String
                         , position :: String
                         , msg :: String }
                         deriving (Show)

instance ToJSON ReqError where
    toJSON reqerr = object [ "tag"      .= ("Err" :: String) -- because of OverloadedStrings
                           , "field"    .= field reqerr
                           , "position" .= position reqerr
                           , "msg"      .= msg reqerr ]

data Response = Resp { stps :: Int
                     , rewseq :: [N.Zipper String String]
                     } deriving Show

instance ToJSON Response where
    toJSON resp = object [ "steps" .= stps resp
                         , "rewriteSeq" .= rewseq resp
                         ]

handleReq :: Request -> Either ReqError Response
handleReq req = do
    prob <- liftErr "main" $ P.parseProb (problem req)
    trm  <- liftErr "term" $ P.parseTerm (P.vars prob) (term req)
    return Resp { stps = 1
                , rewseq = take 500 $ N.normalizeIter (P.annot prob) trm
                }

liftErr :: String -> Either P.ParseError b -> Either ReqError b
liftErr _ (Right r)  = Right r
liftErr fld (Left e) =
    Left ReqError { field    = fld
                  , position = show $ P.errorPos e
                  , msg      = show e }


instance (ToJSON f, ToJSON v) => ToJSON (AnnotElem f v) where
    toJSON (Pos i) = toJSON (i+1)
    toJSON (AnnotRule (NamedRule name _)) = toJSON name

instance (ToJSON f, ToJSON v) => ToJSON (N.Zipper f v) where
    toJSON zipper = object [ "stack"      .= stackJSON (N.stack zipper)
                           , "term"       .= N.term zipper
                           , "annotation" .= N.annotation zipper
                           ]

stackJSON :: (ToJSON f, ToJSON v) => [(Ctxt f v, [AnnotElem f v])] -> Value
stackJSON = toJSON . map go
    where
        go (ctxt, ann) = object [ "context"    .= ctxt
                                , "annotation" .= ann
                                ]
