{-# LANGUAGE DeriveGeneric #-}
{-# LANGUAGE DeriveAnyClass #-}
{-# LANGUAGE OverloadedStrings #-}
-- | This Controller handles the rewrite requests
-- and build the responses.
module Controller.Rewrite
    ( rewriteController
    ) where

import GHC.Generics
import Data.Aeson hiding (Error)
import Web.Scotty as S

import Controller.Strategies (strategies)
import Controller.Orphans() -- Import instances

import AutoStrat.Parsing hiding (term, rule)
import Data.Rewriting.Annotation (getRule)
import Data.Rewriting.Strategies
import Data.Rewriting.Term.Type hiding (map, fold)
import Data.Rewriting.Rule.Type hiding (map)
import Data.Rewriting.Rules (Reduct(..))
import Text.PrettyPrint.ANSI.Leijen

data RewriteRequest = RewReq
    { system  :: String
    , term    :: String
    , strat   :: String
    }
    deriving (Show, Generic, ToJSON, FromJSON)

data RewriteResponse = RewRes
    { steps      :: Int
    , rewriteSeq :: [(Term String String, [Reduct String String String])]
    , nextReq    :: Maybe RewriteRequest
    } | PErr { msg :: String }
    deriving (Generic, ToJSON)

sizeLe (PErr _) i = True
sizeLe (RewRes _ rs _) i = all (< i) $ scanl (\s t -> s + size t) 0 $ map fst rs
    where
        size (Var _) = 0
        size (Fun _ ts) = 1 + sum (map size ts)


-- | The available Strategies form `Controller.Strategies`
strats :: [(String, [Rule String String] -> Term String String -> [(Term String String, [Reduct String String String])])]
strats = map snd strategies

-- | Handles rewrite requests
rewriteController :: ActionM ()
rewriteController = do
    req <- S.jsonData
    let res = case rewriteResponse req of
                Left e  -> PErr (show e)
                Right r -> if r `sizeLe` (10^4) then r else (PErr "Result to Large.")
    S.json res

-- | Parses Request into the variables, TRS and requested term,
-- then constructs response.
rewriteResponse :: RewriteRequest -> Either Err RewriteResponse
rewriteResponse req = do
    prob <- liftErr "MainInput" $ parseProb $ system req
    t    <- liftErr "Term"      $ parseTerm (vars prob) $ term req
    let Just strategy = lookup (strat req) strats
        unnamedRules  = map getRule $ trs prob
        result        = case strat req of
                          "maxstrat" -> take 10 $ strategy unnamedRules t
                          _ -> take 100 $ strategy unnamedRules t
        lastReduct    = fst (last result)
    return RewRes { steps      = length result
                  , rewriteSeq = result
                  , nextReq    = if normalForm unnamedRules lastReduct
                                   then Nothing
                                   else Just req { term = show . pretty $ lastReduct }
                  }


data Err = Error String SourcePos
    deriving Show

liftErr :: String -> Either ParseError b -> Either Err b
liftErr _ (Right a) = Right a
liftErr s (Left e)  = Left $ Error s (errorPos e)

