   {-# LANGUAGE OverloadedStrings #-}
-- | Handles the request for possible strategies
module Controller.Strategies
    ( strategiesController
    , suggestStrategyController
    , strategies
    ) where

import Web.Scotty as S
import Data.Aeson
import Data.Rewriting.Annotation (getRule)
import Data.Rewriting.Strategies
import Data.Rewriting.Properties
import Data.Rewriting.Term.Type (Term(..))
import Data.Rewriting.Rule.Type (Rule(..))
import Data.Rewriting.Rules (Reduct)

import qualified AutoStrat.Parsing as P

-- | Answers with available strategies.
strategiesController :: ActionM ()
strategiesController = S.json names
    where
        names = map (\(name, (ident,_)) -> (name,ident))  strategies


-- | Strategies used in the Application.
--
--  The lookup table is organized as:
--          (Name, (Id, Function))
--    where Name is the name displayed in the Interface,
--          Id is used in the rewrite request,
--          Function is the rewrite strategy used.
strategies ::[(String, (String, [Rule String String] -> Term String String -> [(Term String String, [Reduct String String String])]))]
strategies = [ ("Leftmost-Innermost", ("linner", leftmostInner))
             , ("Leftmost-Outermost", ("louter", leftmostOuter))
             , ("Maximal-Innermost", ("minner", maximalInner))
             , ("Maximal-Outermost", ("mouter", maximalOuter))
             , ("Maximal-Strategy", ("maxstrat", maximalStrat))
             ]

data Request = Req { system :: String
                   }
    deriving (Eq, Show)

instance FromJSON Request where
    parseJSON (Object obj) = Req <$> obj .: "system"

data Suggestion = Strategy [String]
                | NoSuggestion
                | SuggestError String

instance ToJSON Suggestion where
    toJSON (Strategy names) = object [ "tag" .= ("suggestion" :: String)
                                     , "strategies" .= names
                                     ]

    toJSON NoSuggestion = object [ "tag" .= ("NoSuggestion" :: String) ]
    toJSON (SuggestError msg) = object [ "tag" .= ("Err" :: String)
                                       , "msg" .= msg
                                       ]

suggestStrategyController :: ActionM()
suggestStrategyController = do
    req <- S.jsonData
    let res = handleSuggestStrat req
    S.json res

handleSuggestStrat :: Request -> Suggestion
handleSuggestStrat req =
    case go req of
      Right s  -> s
      Left err -> SuggestError (show err)
  where
      go req = do
          prob <- P.parseProb (system req)
          let rules = map getRule (P.trs prob)
          return $ case suggest rules of
                     Strategy [] -> NoSuggestion
                     s -> s

suggest :: [Rule String String] -> Suggestion
suggest rules = Strategy $
    ["Maximal-Outermost"  | isOrthogonal]
 ++ ["Maximal-Strategy"   | isOrthogonal]
 ++ ["Leftmost-Outermost" | isLeftNormal && isOrthogonal]
     where
         isOrthogonal = orthogonal rules
         isLeftNormal = leftNormal rules

